from specifications.merkletree import MerkleTreeProtocol, check_module
from context import merkleasy


check_module(merkleasy, {merkleasy.Tree: MerkleTreeProtocol})
print('specification test complete')
